/*
 * DNSReader - reads output from the various writers
 *
 * Copyright 2007 (c) Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/queue.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <err.h>
#include <fcntl.h>
#include <netdb.h>
#include <unistd.h>
#include <assert.h>
#include <ctype.h>

#include <event.h>
#include <evdns.h>
#include <dnet.h>

#include "dnsdata.gen.h"
#include "utils.h"
#include "fpdns.h"
#include "dnscore.h"
#include "exclude.h"

#define DNSSCAN_TAG_REQUEST	0x01
#define CHECKPOINT_TAG		0x02
#define DNSRESPONDER_TAG	0xae

static const char *search_string = NULL;
static const char *match_query = NULL;
static const char *match_resolver = NULL;

static int filter_payload = 0;
static int filter_open_recursive = 0;
static int filter_liars = 0;
static int show_date = 0;
static int show_payload = 0;
static int show_chaos_reply = 0;
static int show_fingerprint = 0;
static int show_http = 0;
static int filter_forward = 1;
static int show_date_in_usec = 0;

static char *
escape_payload(uint8_t *data, uint32_t datalen)
{
	char *payload, *p;
	int i, count = 0;

	for (i = 0; i < datalen; ++i) {
		if (isprint(data[i]) || isspace(data[i]))
			count += 1;
		else
			count += 3;
	}

	payload = malloc(count + 1);
	assert(payload != NULL);

	p = payload;
	for (i = 0; i < datalen; ++i) {
		if (isprint(data[i]) || isspace(data[i])) {
			*p++ = data[i];
		} else {
			static char buf[4];

			snprintf(buf, sizeof(buf),
			    "\\%02x", data[i]);

			memcpy(p, buf, 3);
			p += 3;
		}
	}
	*p = '\0';

	return (payload);
}

int
read_dnsresponder(struct evbuffer *buf)
{
	struct timeval tv;
	struct DNSResponder *dr;
	uint8_t *data = NULL;
	uint32_t data_len;
	char *query;

	dr = DNSResponder_new();
	assert(dr != NULL);

	if (evtag_unmarshal_DNSResponder(buf, DNSRESPONDER_TAG, dr) == -1)
		return (-1);

	if (EVTAG_GET(dr, remote_addr, &data, &data_len) == -1)
		return (-1);
	if (data_len != sizeof(struct addr))
		return (-1);

	/* only output if the resolver matches */
	if (match_resolver &&
	    strcmp(addr_ntoa((struct addr *)data), match_resolver))
		return (0);

	if (EVTAG_GET(dr, tv_sec, (uint32_t *)&tv.tv_sec) == -1)
		return (-1);
	if (EVTAG_GET(dr, tv_sec, (uint32_t *)&tv.tv_usec) == -1)
		return (-1);

	if (show_date) {
		static char date[50];
		time_t when_in_sec;
		struct tm when_as_tm;

		when_in_sec = tv.tv_sec;
		localtime_r(&when_in_sec, &when_as_tm);
		strftime(date, sizeof(date),
		    "%Y-%m-%d-%H:%M:%S", &when_as_tm);

		fprintf(stdout, "%s ", date);
	} else if (show_date_in_usec) {
		fprintf(stdout, "%llu ",
		    (uint64_t)tv.tv_sec * 1000000L + tv.tv_usec);
	}

	if (EVTAG_GET(dr, query, &query) == -1)
		return (-1);
	
	if (filter_forward) {
		struct addr forward;
		char *dot = strchr(query, '.');
		if (dot == NULL)
			return (0);

		*dot = '\0';

		/* 32-bit IPv4 address encoded as hex */
		if (strlen(query) != 4 * 2)
			return (0);

		forward.addr_type = ADDR_TYPE_IP;
		forward.addr_bits = IP_ADDR_BITS;
		if (ascii_to_bin(query, (uint8_t *)&forward.addr_ip,
			sizeof(forward.addr_ip)) == -1)
			return (0);

		query = addr_ntoa(&forward);
	}

	fprintf(stdout, "%s %s\n", addr_ntoa((struct addr *)data), query);

	DNSResponder_free(dr);

	return (0);
}

int
read_dnsrequest(struct evbuffer *buf)
{
	struct timeval tv;
	struct DNSRequest *dr;
	uint8_t *data = NULL;
	uint32_t data_len;

	dr = DNSRequest_new();
	assert(dr != NULL);
	if (evtag_unmarshal_DNSRequest(buf, DNSSCAN_TAG_REQUEST, dr) == -1)
		return (-1);

	if (filter_open_recursive) {
		struct dns_reply reply;

		if (EVTAG_GET(dr, payload, &data, &data_len) == -1)
			return (-1);
		if (dnscore_reply_parse(EVDNS_TYPE_A, &reply,
			data, data_len) != 0)
			return (-1);

		/* ignore all resolvers that did not return an address */
		if (!reply.data.a.addrcount)
			return (0);
	}

	if (EVTAG_GET(dr, remote_addr, &data, &data_len) == -1)
		return (-1);
	if (data_len != sizeof(struct addr))
		return (-1);

	/* only output if the resolver matches */
	if (match_resolver &&
	    strcmp(addr_ntoa((struct addr *)data), match_resolver))
		return (0);

	if (search_string) {
		uint8_t *body;
		uint32_t bodylen;

		if (!EVTAG_HAS(dr, http_body_payload))
			return (0);

		if (EVTAG_GET(dr, http_body_payload, &body, &bodylen) == -1)
			return (-1);
		
		if (strnstr((char *)body, search_string, bodylen) == 0) {
			return (0);
		}
	}

	if (EVTAG_GET(dr, tv_sec, (uint32_t *)&tv.tv_sec) == -1)
		return (-1);
	if (EVTAG_GET(dr, tv_sec, (uint32_t *)&tv.tv_usec) == -1)
		return (-1);

	if (show_date) {
		static char date[50];
		time_t when_in_sec;
		struct tm when_as_tm;

		when_in_sec = tv.tv_sec;
		localtime_r(&when_in_sec, &when_as_tm);
		strftime(date, sizeof(date),
		    "%Y-%m-%d-%H:%M:%S", &when_as_tm);

		fprintf(stdout, "%s ", date);
	} else if (show_date_in_usec) {
		fprintf(stdout, "%llu ",
		    (uint64_t)tv.tv_sec * 1000000L + tv.tv_usec);
	}

	fprintf(stdout, "%s", addr_ntoa((struct addr *)data));

	if (show_payload) {
		char *buf;
		if (EVTAG_GET(dr, payload, &data, &data_len) == -1)
			return (-1);

		buf = malloc(data_len * 2 + 1);
		assert(buf != NULL);
		bin_to_ascii(data, data_len, buf, data_len * 2 + 1);

		fprintf(stdout, " %s", buf);
		free(buf);
	}

	if (show_fingerprint) {
		static char ascii[32];
		const char *fp = "_";

		if (EVTAG_GET(dr, payload, &data, &data_len) == -1)
			return (-1);
		if (fpdns_print(data, data_len,
			ascii, sizeof(ascii)) == -1)
			return (-1);
		fp = ascii;
		fprintf(stdout, " %s", fp);
		
		fp = "_";
		if (EVTAG_HAS(dr, fpdns_iquery))
			EVTAG_GET(dr, fpdns_iquery, (char **)&fp);
		fprintf(stdout, " %s", fp);

		fp = "_";
		if (EVTAG_HAS(dr, fpdns_tc))
			EVTAG_GET(dr, fpdns_tc, (char **)&fp);
		fprintf(stdout, " %s", fp);

		fp = "_";
		if (EVTAG_HAS(dr, chaos_sent))
			fp = "t";
		if (EVTAG_HAS(dr, chaos_payload)) {
			if (EVTAG_GET(dr, chaos_payload,
				&data, &data_len) == -1)
				return (-1);

			if (fpdns_print(data, data_len,
				ascii, sizeof(ascii)) == -1)
				return (-1);
			fp = ascii;
		}
		fprintf(stdout, " %s", fp);
	}

	if (show_chaos_reply) {
		const char *reply = "(not sent)";
		if (EVTAG_HAS(dr, chaos_sent))
			reply = "(timeout)";
		if (EVTAG_HAS(dr, chaos_reply))
			EVTAG_GET(dr, chaos_reply, (char **)&reply);
		if (!strlen(reply))
			reply = "(empty)";
		fprintf(stdout, " %s", reply);
	}

	if (show_http) {
		const char *http = "(no answer)";
		unsigned int http_len = -1;

		if (EVTAG_HAS(dr, http_body_payload))
			http = "(no header)";
		if (EVTAG_HAS(dr, fphttp_server))
			EVTAG_GET(dr, fphttp_server, (char **)&http);

		if (EVTAG_HAS(dr, fphttp_body_len)) {
			if (EVTAG_GET(dr, fphttp_body_len, &http_len) == -1)
				return (-1);
		}

		fprintf(stdout, " %s (%d)", http, http_len);
	}

	if (show_payload && EVTAG_HAS(dr, http_body_payload)) {
		char *buf;
		if (EVTAG_GET(dr, http_body_payload, &data, &data_len) == -1)
			return (-1);

		buf = escape_payload(data, data_len);
		fprintf(stdout, "\n======\n%s\n======\n", buf);
		free(buf);
	}

	DNSRequest_free(dr);

	fprintf(stdout, "\n");

	return (0);
}

int
print_server_reply(struct HTTPServerReply *reply)
{
	unsigned int response_code;
	char *response_line;
	int i;

	if (EVTAG_GET(reply, response_code, &response_code) == -1)
		return (-1);
	if (EVTAG_GET(reply, response_string, &response_line) == -1)
		return (-1);

	fprintf(stdout, "\n\tHTTP %d %s\n", response_code, response_line);

	for (i = 0; i < EVTAG_LEN(reply, headers); ++i) {
		struct HTTPHeader *header;
		char *key, *value;

		if (EVTAG_GET(reply, headers, i, &header) == -1)
			return (-1);
		if (EVTAG_GET(header, key, &key) == -1)
			return (-1);
		if (EVTAG_GET(header, value, &value) == -1)
			return (-1);
		fprintf(stdout, "\t%s: %s\n", key, value);

	}

	if (show_payload && EVTAG_HAS(reply, body_content)) {
		char *payload;
		uint8_t *data;
		uint32_t data_len;

		if (EVTAG_GET(reply, body_content, &data, &data_len) == -1)
			return (-1);

		payload = escape_payload(data, data_len);

		fprintf(stdout, "\n======\n%s\n======\n", payload);

		free(payload);
	}

	return (0);
}

void
read_data(fd)
{
	struct evbuffer *buf = evbuffer_new();
	int count = 0;
	assert(buf != NULL);

	while (evbuffer_read(buf, fd, -1) > 0) {
		while (EVBUFFER_LENGTH(buf) > 4) {
			uint32_t tag;
			uint32_t length;

			if (evtag_peek_length(buf, &length) == -1)
				goto error;

			if (EVBUFFER_LENGTH(buf) < length)
				break;

			if (evtag_peek(buf, &tag) == -1)
				goto error;

			count++;

			switch (tag) {
			case DNSSCAN_TAG_REQUEST:
				if (read_dnsrequest(buf) == -1) {
					/* skip bad entries */
					fprintf(stderr,
					    "Skipping corrupt record\n");
					continue;
				}
				break;
			case DNSRESPONDER_TAG:
				if (read_dnsresponder(buf) == -1) {
					/* skip bad entries */
					fprintf(stderr,
					    "Skipping corrupt record\n");
					continue;
				}
				break;
			case CHECKPOINT_TAG:
				fprintf(stderr, "Not implemented\n");
				exit(1);
			default:
				fprintf(stderr, "Unknown data set: %d\n", tag);
				exit(1);
				break;
			}
		}
	}

	fprintf(stderr, "... read %d records\n", count);

	return;

error:
	fprintf(stderr, "Corrupted input at record %d\n", count);
}

static void
usage(const char *name)
{
	fprintf(stderr, 
	    "%s: [-D] <filenames...>\n\n"
	    "\t-o	      only show open recursive resolvers\n"
	    "\t-D             show date data was discovered\n"
	    "\t-P             show DNS payload in hex\n"
	    "\t-C	      show DNS Chaos reply\n"
	    "\t-F	      show fingerprint results\n"
	    "\t-m <ipaddr>    show records only matching this resolvers\n"
	    "\t-q <query>     show records only matching this query\n"
	    "\t-s <search>    show records that match this search string\n",
	    name);
}

static void
read_with_unpack(const char *filename)
{
	int fd = -1;
	FILE *fp = NULL;
	static char command[1024];
	char *unpacked_name = strdup(filename);
	char *lastdot, *extension;
	assert(unpacked_name != NULL);

	command[0] = '\0';
	lastdot = strrchr(unpacked_name, '.');
	extension = lastdot != NULL ? lastdot + 1 : NULL;

	if (extension) {
		if (strcmp(extension, "bz2") == 0) {
			snprintf(command, sizeof(command),
			    "bzip2 -dc \"%s\"", unpacked_name);
		} else if (strcmp(extension, "gz") == 0) {
			snprintf(command, sizeof(command),
			    "gzip -dc \"%s\"", unpacked_name);
		}
	}

	if (strlen(command) > 0) {
		*lastdot = '\0';
		fp = popen(command, "r");
		if (fp == NULL)
			err(1, "cannot read \"%s\"", filename);
		fd = fileno(fp);
		assert(fd != -1);
	} else {
		fd = open(filename, O_RDONLY, 0);
		if (fd == -1)
			err(1, "cannot read \"%s\"", filename);
	}

	read_data(fd);

	free(unpacked_name);

	if (fp != NULL) {
		pclose(fp);
	} else {
		assert(fd != -1);
		close(fd);
	}
}

int
main(int argc, char **argv)
{
	const char *name = argv[0];
	int ch;

	event_init();
	evtag_init();

	while ((ch = getopt(argc, argv, "ploHFCVPUDm:q:s:")) != -1)
		switch(ch) {
		case 's':
			search_string = optarg;
			break;
		case 'q':
			match_query = optarg;
			break;
		case 'm':
			match_resolver = optarg;
			break;
		case 'p':
			filter_payload = 1;
			break;
		case 'o':
			filter_open_recursive = 1;
			break;
		case 'l':
			filter_liars = 1;
			break;
		case 'H':
			show_http = 1;
			break;
		case 'F':
			show_fingerprint = 1;
			break;
		case 'C':
			show_chaos_reply = 1;
			break;
		case 'D':
			show_date = 1;
			break;
		case 'U':
			show_date_in_usec = 1;
			break;
		case 'P':
			show_payload = 1;
			break;
		case 'V':
			fprintf(stderr, "DNSResponder %s\n", VERSION);
			exit(0);
		case 'h':
		default:
			usage(name);
			exit(1);
		}

	argc -= optind;
	argv += optind;

	if (!argc) {
		fprintf(stderr, "... No input specified ... Using stdin\n");

		read_data(fileno(stdin));
		return (0);
	}

	while (argc) {
		const char *filename = *argv;

		fprintf(stderr, "...opening %s\n", filename);
		read_with_unpack(filename);
		
		argv++;
		argc--;
	}

	return (0);
}
