/*
 * Copyright 2007 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/queue.h>
#include <sys/time.h>

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <strings.h>
#include <fcntl.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <time.h>
#include <string.h>

#include <event.h>

#include "evutils.h"

/* Event Counter Code */

static void
event_counter_add_slot(struct event_slot *slot, uint32_t count)
{
	struct timeval tv, tv_before;
	struct event_single_slot *ss = &slot->slots[slot->cur_pos];
	
	gettimeofday(&tv, NULL);
	timersub(&tv, &slot->interval, &tv_before);

	if (timercmp(&tv_before, &ss->tv, <=)) {
		ss->count += count;
		return;
	}

	/* if this is not the first use, we need to rotate the slot */
	if (ss->tv.tv_sec || ss->tv.tv_usec) {
		if (++slot->cur_pos >= EVENT_SLOTS)
			slot->cur_pos = 0;
		ss = &slot->slots[slot->cur_pos];
	}

	ss->tv = tv;
	ss->count = count;
}

struct event_counter *
event_counter_new(const char *name, int num_entries, struct timeval *intervals)
{
	struct event_counter *ctr = calloc(1, sizeof(struct event_counter));
	int i;
	assert(ctr != NULL);

	ctr->num_slots = num_entries;
	ctr->slots = calloc(num_entries, sizeof(struct event_slot));
	assert(ctr->slots != NULL);

	ctr->name = strdup(name);
	assert(ctr->name != NULL);

	for (i = 0; i < num_entries; ++i) {
		struct event_slot *slot = &ctr->slots[i];
		slot->interval = intervals[i];

		slot->threshold.tv_sec = slot->interval.tv_sec * EVENT_SLOTS;
		slot->threshold.tv_usec = slot->interval.tv_usec * EVENT_SLOTS;
		slot->threshold.tv_sec += slot->threshold.tv_usec / 1000000L;
		slot->threshold.tv_usec %= 1000000L;
	}

	return (ctr);
}

void
event_counter_free(struct event_counter *ctr)
{
	free(ctr->slots);
	free(ctr->name);
	free(ctr);
}

void
event_counter_add(struct event_counter *ctr, uint32_t count)
{
	int i;
	ctr->total += count;

	for (i = 0; i < ctr->num_slots; ++i)
		event_counter_add_slot(&ctr->slots[i], count);
}

static void
event_counter_get_slot(struct event_slot *slot, uint32_t *pcount)
{
	struct timeval tv;
	uint32_t count = 0;
	int i = slot->cur_pos;
	
	gettimeofday(&tv, NULL);
	timersub(&tv, &slot->threshold, &tv);

	/*
	 * the time of uninitialized slots is 0.0, so the compare is
	 * going to terminate on them.
	 */
	do {
		struct event_single_slot *ess = &slot->slots[i--];
		if (timercmp(&ess->tv, &tv, <))
			break;
		count += ess->count;
		if (i < 0)
			i = EVENT_SLOTS - 1;
	} while (i != slot->cur_pos);
	
	*pcount = count;
}

void
event_counter_get(struct event_counter *ctr, uint32_t *counts)
{
	int i;
	for (i = 0; i < ctr->num_slots; ++i)
		event_counter_get_slot(&ctr->slots[i], counts++);
	*counts = ctr->total;
}

/* Histogram Code */
struct event_histogram *
event_histogram_new(int num_slots)
{
	struct event_histogram *ehg;
	ehg = calloc(1, sizeof(struct event_histogram));
	assert(ehg != NULL);

	ehg->num_slots = num_slots;
	ehg->counters = calloc(num_slots, sizeof(struct event_counter *));

	return (ehg);
}

void
event_histogram_add(struct event_histogram *ehg,
    const char *name, uint32_t count)
{
	struct event_counter *counter = NULL;
	int i;

	for (i = 0; i < ehg->used_slots; i++) {
		counter = ehg->counters[i];
		if (strcmp(name, counter->name) == 0)
			break;
	}

	if (i == ehg->used_slots) {
		struct timeval tvs[2] = { { 1, 0 }, { 60, 0 } };
		/* we need to create a new one */
		if (ehg->used_slots == ehg->num_slots) {
			/* no sizing at the moment */
			return;
		}
		counter = event_counter_new(name, 2, tvs);
		ehg->counters[ehg->used_slots++] = counter;
	}

	event_counter_add(counter, count);
}

static int
counter_compare(const void *a, const void *b)
{
	struct event_counter *cnt_a = *(struct event_counter **)a;
	struct event_counter *cnt_b = *(struct event_counter **)b;

	if (cnt_a->total < cnt_b->total)
		return (1);
	else if (cnt_a->total > cnt_b->total)
		return (-1);
	
	return (0);
}

void
event_histogram_as_html(struct evbuffer *buf, const char *css_class,
    struct event_histogram *ehg)
{
	uint32_t counts[3];
	int i;
	
	if (!ehg->used_slots)
		return;

	/* sort the slots by frequency */
	qsort(ehg->counters, ehg->used_slots, sizeof(struct event_counter *),
	    counter_compare);

	evbuffer_add_printf(buf, 
	    "<table class=\"%s\">"
	    "<tr>"
	    "<th>Name</th><th>Minute</th><th>Hour</th><th>Total</th>"
	    "</tr>\n", css_class);

	for (i = 0; i < ehg->used_slots; i++) {
		struct event_counter *cnt = ehg->counters[i];
		event_counter_get(cnt, counts);
		evbuffer_add_printf(buf,
		    "<tr><td>%s</td><td>%d</td><td>%d</td><td>%d</td></tr>\n",
		    cnt->name, counts[0], counts[1], counts[2]);
	}
	

	evbuffer_add_printf(buf, "</table>\n");
}

/* History Code */

struct history_buffer *
history_buffer_new(int max_entries)
{
	struct history_buffer *hist = calloc(1, sizeof(struct history_buffer));
	assert(hist != NULL);

	TAILQ_INIT(&hist->entries);
	hist->max_entries = max_entries;

	return (hist);
}

void
history_enter_buffer(struct history_buffer *hist, struct evbuffer *buf)
{
	struct history_entry *entry;

	if (hist->num_entries >= hist->max_entries) {
		entry = TAILQ_FIRST(&hist->entries);
		TAILQ_REMOVE(&hist->entries, entry, next);
		evbuffer_free(entry->ev_buf);
		entry->ev_buf = NULL;
	} else {
		hist->num_entries++;
		entry = calloc(1, sizeof(struct history_entry));
		assert(entry != NULL);
	}

	entry->ev_buf = buf;
	gettimeofday(&entry->tv, NULL);

	TAILQ_INSERT_TAIL(&hist->entries, entry, next);
}

void
history_to_html(struct evbuffer *output, struct history_buffer *hist)
{
	struct history_entry *entry;

	/* check if we have any work todo */
	if (TAILQ_FIRST(&hist->entries) == NULL)
		return;

	evbuffer_add_printf(output, "<table>\n");
	evbuffer_add_printf(output, "<th>Time</th><th>Entry</th>\n");

	TAILQ_FOREACH(entry, &hist->entries, next) {
		static char date[50];
		time_t when_in_sec = entry->tv.tv_sec;
		struct tm when_as_tm;

		localtime_r(&when_in_sec, &when_as_tm);
		strftime(date, sizeof(date),
		    "%Y-%m-%d-%H:%M:%S", &when_as_tm);

		evbuffer_add_printf(output, "<tr><td>%s</td><td>", date);
		evbuffer_add(output,
		    EVBUFFER_DATA(entry->ev_buf),
		    EVBUFFER_LENGTH(entry->ev_buf));
		evbuffer_add_printf(output, "</td></tr>\n");
	}

	evbuffer_add_printf(output, "</table>\n");
}

/* misc utility */

void
event_seconds_as_eta(time_t seconds, char *output, size_t out_len)
{
	snprintf(output, out_len,
	    "%d days %02d:%02d:%02d",
	    (int)(seconds / (24 * 3600)),
	    (int)((seconds / 3600) % 24),
	    (int)((seconds / 60) % 60),
	    (int)(seconds % 60));
}
